#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import logging
from enum import Enum, auto
from typing import TYPE_CHECKING, List, Optional

from qgis.PyQt.QtCore import QObject, QThread, QTimer, pyqtSignal, pyqtSlot
from qgis_plugin_tools.tools.i18n import tr

from quality_result_gui.api.quality_api_client import (
    QualityResultClient,
    QualityResultClientError,
    QualityResultServerError,
)

if TYPE_CHECKING:
    from quality_result_gui.api.types.quality_error import QualityErrorsByPriority

BACKGROUND_POLL_INTERVAL = 10 * 1000

LOGGER = logging.getLogger(__name__)


class CheckStatus(Enum):
    CHECKING = auto()
    RESULT_ONGOING = auto()
    RESULT_UPDATED = auto()
    RESULT_FAILED = auto()


CHECK_STATUS_LABELS = {
    CheckStatus.CHECKING: tr("Checking for quality result updates"),
    CheckStatus.RESULT_ONGOING: tr("Quality check is in progress"),
    CheckStatus.RESULT_FAILED: tr("Quality result update failed"),
    CheckStatus.RESULT_UPDATED: tr("Quality results are up to date"),
}


class PollingWorker(QObject):
    status_changed = pyqtSignal(CheckStatus)
    results_received = pyqtSignal(list)

    def __init__(
        self,
        api_client: QualityResultClient,
        parent: Optional[QObject] = None,
        poll_interval: int = BACKGROUND_POLL_INTERVAL,
    ) -> None:
        super().__init__(parent)
        self._timer: Optional[QTimer] = None
        self._poll_interval = poll_interval
        self._api_client = api_client

    @pyqtSlot()
    def start(self) -> None:
        if self._timer is not None:
            self._timer.stop()

        # Run every x seconds
        self._timer = QTimer(self)
        self._timer.timeout.connect(self._check_api)
        self._timer.start(self._poll_interval)

        # For the first time run right now since timer triggers it after x seconds
        self._check_api()

    @pyqtSlot()
    def _check_api(self) -> None:
        self.status_changed.emit(CheckStatus.CHECKING)

        try:
            results = self._api_client.get_results()

            if results is None:
                self.status_changed.emit(CheckStatus.RESULT_ONGOING)
            else:
                self.status_changed.emit(CheckStatus.RESULT_UPDATED)
                self.results_received.emit(results)

        except (QualityResultClientError, QualityResultServerError) as e:
            LOGGER.warning(
                f"failed to check quality results api: {str(e)}", stack_info=True
            )
            self.status_changed.emit(CheckStatus.RESULT_FAILED)


class BackgroundQualityResultsFetcher(QObject):
    status_changed = pyqtSignal(CheckStatus)
    results_received = pyqtSignal(list)

    def __init__(
        self,
        api_client: QualityResultClient,
        parent: Optional[QObject] = None,
        poll_interval: int = BACKGROUND_POLL_INTERVAL,
    ) -> None:
        super().__init__(parent)
        self._thread: Optional[QThread] = None
        self._worker: Optional[PollingWorker] = None
        self._poll_interval = poll_interval
        self._api_client = api_client

    @pyqtSlot(bool)
    def set_checks_enabled(self, enabled: bool) -> None:
        if enabled:
            self.start()
        else:
            self.stop()

    @pyqtSlot(CheckStatus)
    def _worker_status_changed(self, status: CheckStatus) -> None:
        self.status_changed.emit(status)

    @pyqtSlot(list)
    def _worker_results_received(
        self, results: List["QualityErrorsByPriority"]
    ) -> None:
        self.results_received.emit(results)

    @pyqtSlot()
    def start(self) -> None:
        self.stop()
        self._thread = QThread(self)
        self._worker = PollingWorker(self._api_client, None, self._poll_interval)
        self._worker.moveToThread(self._thread)

        self._thread.finished.connect(self._worker.deleteLater)
        self._thread.started.connect(self._worker.start)

        self._worker.status_changed.connect(self._worker_status_changed)
        self._worker.results_received.connect(self._worker_results_received)

        self._thread.start()

    @pyqtSlot()
    def stop(self) -> None:
        if self._thread and self._thread.isRunning():
            self._thread.quit()
            self._thread.wait()
        self._thread = None



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Optional

from qgis.core import QgsGeometry
from qgis_plugin_tools.tools.i18n import tr


class QualityErrorType(Enum):
    ATTRIBUTE = 1
    GEOMETRY = 2
    TOPOLOGY = 3
    CONTINUITY = 4


ERROR_TYPE_LABEL = {
    QualityErrorType.ATTRIBUTE: tr("Attribute error"),
    QualityErrorType.GEOMETRY: tr("Geometry error"),
    QualityErrorType.TOPOLOGY: tr("Topology error"),
    QualityErrorType.CONTINUITY: tr("Continuity error"),
}


class QualityErrorPriority(Enum):
    FATAL = 1
    WARNING = 2
    INFO = 3


ERROR_PRIORITY_LABEL = {
    QualityErrorPriority.FATAL: tr("Fatal"),
    QualityErrorPriority.WARNING: tr("Warning"),
    QualityErrorPriority.INFO: tr("Info"),
}


@dataclass
class QualityError:
    priority: QualityErrorPriority
    feature_type: str
    feature_id: str
    error_id: int
    unique_identifier: str
    error_type: QualityErrorType
    attribute_name: Optional[str]
    error_description: str
    error_extra_info: str
    geometry: QgsGeometry
    is_user_processed: bool

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)


@dataclass
class QualityErrorsByFeature:
    feature_type: str
    feature_id: str
    errors: List[QualityError]


@dataclass
class QualityErrorsByFeatureType:
    feature_type: str
    errors: List[QualityErrorsByFeature]

    def get_all_errors(self) -> List[QualityError]:
        errors = [errors_by_feature.errors for errors_by_feature in self.errors]
        return [item for sub_list in errors for item in sub_list]


@dataclass
class QualityErrorsByPriority:
    priority: QualityErrorPriority
    errors: List[QualityErrorsByFeatureType]

    def get_all_errors(self) -> List[QualityError]:
        errors = [
            errors_by_feature_type.get_all_errors()
            for errors_by_feature_type in self.errors
        ]
        return [item for sub_list in errors for item in sub_list]



#  Copyright (C) 2022-2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from types import GeneratorType
from typing import TYPE_CHECKING, Generator, Optional, cast

from qgis.PyQt.QtCore import QAbstractItemModel, QModelIndex, Qt, QVariant, pyqtSignal
from qgis.PyQt.QtWidgets import QTreeView, QWidget
from qgis_plugin_tools.tools.decorations import log_if_fails

from quality_result_gui import SelectionType
from quality_result_gui.api.types.quality_error import QualityError
from quality_result_gui.quality_errors_tree_model import (
    ErrorDataType,
    ModelColumn,
    QualityErrorTreeItemType,
)

if TYPE_CHECKING:
    from qgis.PyQt.QtGui import QMouseEvent

TREE_VIEW_STYLE = """
        QTreeView::item:hover, QTreeView::item:selected:active,
        QTreeView::item:selected, QTreeView::branch:selected:active,
        QTreeView::branch:hover {
            background: #f5f5f5;
            border-bottom: 1px solid #f5f5f5;
            color: black;
        }
        QTreeView::item {
            border-bottom: 1px solid #f5f5f5;
            background: white;
        }
        """


class QualityErrorTreeView(QTreeView):
    quality_error_selected = pyqtSignal(QualityError, SelectionType)

    errors_inserted = pyqtSignal(GeneratorType)
    errors_removed = pyqtSignal(GeneratorType)

    def __init__(
        self,
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent)

        self.source_button_event: Optional["QMouseEvent"] = None

        self.setColumnWidth(ModelColumn.TYPE_OR_ID.value, 150)
        self.setIndentation(10)
        self.setUniformRowHeights(True)

        self.setStyleSheet(TREE_VIEW_STYLE)

    def setModel(  # noqa: N802 (override qt method)
        self, model: Optional[QAbstractItemModel]
    ) -> None:
        super().setModel(model)
        if not model:
            return

        self.selectionModel().currentChanged.connect(self._on_current_item_changed)

        model.rowsInserted.connect(self._on_model_rows_inserted)
        model.rowsAboutToBeRemoved.connect(self._on_rows_about_to_be_removed)

    def mousePressEvent(  # noqa: N802 (override qt method)
        self, event: "QMouseEvent"
    ) -> None:
        self.source_button_event = event
        # Calling super will trigger currentChanged if row clicked
        super().mousePressEvent(event)

        self.source_button_event = None

    def _on_model_rows_inserted(
        self, parent: QModelIndex, first: int, last: int
    ) -> None:
        self.expandRecursively(parent)

        for i in range(first, last + 1):
            index = self.model().index(i, 0, parent)

            # Update visualized errors
            new_errors_to_visualize = self._get_quality_errors_from_index(index)
            self.errors_inserted.emit(new_errors_to_visualize)

    def _on_rows_about_to_be_removed(
        self, parent: QModelIndex, first: int, last: int
    ) -> None:
        for i in range(first, last + 1):
            index = self.model().index(i, 0, parent)

            # Update visualized errors
            errors_to_remove = self._get_quality_errors_from_index(index)
            self.errors_removed.emit(errors_to_remove)

    @log_if_fails
    def _on_current_item_changed(
        self, current_index: QModelIndex, previous_index: QModelIndex
    ) -> None:
        quality_error = self._get_quality_error_from_row(current_index)

        if quality_error is None:
            return

        selection_mode = (
            SelectionType.RightClick
            if (
                self.source_button_event is not None
                and self.source_button_event.button() == Qt.MouseButton.RightButton
            )
            else SelectionType.Other
        )

        self.quality_error_selected.emit(quality_error, selection_mode)

    def _get_quality_errors_from_index(
        self, index: QModelIndex
    ) -> Generator[QualityError, None, None]:
        """Get quality errors recursively from index."""

        if not index.isValid():
            return

        row_count = self.model().rowCount(index)
        if row_count == 0:
            # Index may now be at quality error row, which has never any children
            error = self._get_quality_error_from_row(index)
            if error is not None:
                yield error
            return
        else:
            for i in range(row_count):
                yield from self._get_quality_errors_from_index(index.child(i, 0))

    def _get_quality_error_from_row(
        self,
        row_index: QModelIndex,
    ) -> Optional[QualityError]:
        data = row_index.data(Qt.UserRole)

        if not QVariant(data).isValid():
            return None

        item_type, item_data = cast(ErrorDataType, data)
        # Single quality error
        if item_type == QualityErrorTreeItemType.ERROR:
            return cast(QualityError, item_data)
        return None



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.


import logging
from typing import TYPE_CHECKING, Optional

from qgis.PyQt.QtWidgets import QAction, QMenu, QWidget
from qgis_plugin_tools.tools.i18n import tr

if TYPE_CHECKING:
    from quality_result_gui.quality_errors_filters import FilterMenu

LOGGER = logging.getLogger(__name__)


class QualityErrorsTreeFilterMenu(QMenu):
    """
    Menu for filtering rows in feature tree view inside quality errors dialog.


    """

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.reset_separator = self.addSeparator()
        self.reset_action = QAction(tr("Reset filters"), self)
        self.addAction(self.reset_action)

    def add_filter_menu(self, menu: "FilterMenu") -> None:
        """Adds a menu into this menu as a sub menu

        Args:
            menu (FilterMenu): menu to be added
        """

        self.insertMenu(self.reset_separator, menu)  # insert before the separator
        self.reset_action.triggered.connect(menu._select_all)

    def is_any_filter_active(self, menu: Optional[QMenu] = None) -> bool:
        """Checks if any of menu's checkbox are unchecked

        Recursively travels the menu and its submenus to check if any filter is active

        Args:
            parent (Optional[QMenu], optional): Menu to be checked for. If None, checks
              this menu. Defaults to None.

        Returns:
            bool: Returns True if any of checkboxes in this or any sub menu is checked.
              False otherwise
        """

        menu = menu or self
        for action in menu.actions():
            if action.isCheckable() and action.isChecked() is False:
                return True
            sub_menu = action.menu()
            if sub_menu and self.is_any_filter_active(sub_menu):
                return True
        return False



#  Copyright (C) 2022-2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Union

from qgis.core import (
    QgsFillSymbol,
    QgsLineSymbol,
    QgsMarkerSymbol,
    QgsSymbol,
    QgsWkbTypes,
)
from qgis.PyQt.QtGui import QColor


class BaseSymbol(ABC):
    @abstractmethod
    def to_qgs_symbol(self, geometry_type: QgsWkbTypes.GeometryType) -> QgsSymbol:
        """
        Implement logic to generate the relevant symbol for the given geometry type.
        """
        raise NotImplementedError()

    @staticmethod
    def color_to_rgba_string(color: Union[str, QColor]) -> str:
        return ",".join(map(str, QColor(color).getRgb()))


class ErrorSymbol(BaseSymbol, ABC):
    def to_qgs_symbol(
        self, geometry_type: QgsWkbTypes.GeometryType, highlighted: bool = False
    ) -> QgsSymbol:
        if geometry_type == QgsWkbTypes.PolygonGeometry:
            return self._get_polygon_symbol(highlighted)
        if geometry_type == QgsWkbTypes.LineGeometry:
            return self._get_line_symbol(highlighted)
        if geometry_type == QgsWkbTypes.PointGeometry:
            return self._get_point_symbol(highlighted)
        raise NotImplementedError()

    @abstractmethod
    def _get_polygon_symbol(self, highlighted: bool) -> QgsFillSymbol:
        raise NotImplementedError()

    @abstractmethod
    def _get_line_symbol(self, highlighted: bool) -> QgsLineSymbol:
        raise NotImplementedError()

    @abstractmethod
    def _get_point_symbol(self, highlighted: bool) -> QgsMarkerSymbol:
        raise NotImplementedError()



#  Copyright (C) 2022-2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Union

from qgis.core import (
    QgsAnnotationLayer,
    QgsAnnotationLineItem,
    QgsAnnotationMarkerItem,
    QgsAnnotationPolygonItem,
    QgsGeometry,
    QgsLineString,
    QgsPoint,
    QgsPolygon,
    QgsProject,
    QgsWkbTypes,
)
from qgis_plugin_tools.tools.exceptions import QgsPluginException
from qgis_plugin_tools.tools.i18n import tr

from quality_result_gui.configuration import QualityLayerStyleConfig
from quality_result_gui.style.default_style import DefaultErrorSymbol

if TYPE_CHECKING:
    from quality_result_gui.api.types.quality_error import QualityError
    from quality_result_gui.style.quality_layer_error_symbol import ErrorSymbol

LOGGER = logging.getLogger(__name__)


class LayerException(QgsPluginException):
    pass


class DefaultStyleConfig(QualityLayerStyleConfig):
    def create_error_symbol(self, quality_error: "QualityError") -> "ErrorSymbol":
        return DefaultErrorSymbol(quality_error)


class QualityErrorLayer:
    LAYER_ID = "quality-errors"
    LAYER_ID_PROPERTY = "quality-result-gui-layer"

    def __init__(self) -> None:
        self._annotation_ids: Dict[str, List[str]] = {}
        self.style: "QualityLayerStyleConfig" = DefaultStyleConfig()

    @property
    def annotation_layer(self) -> QgsAnnotationLayer:
        return self.get_annotation_layer()

    def override_style(
        self,
        style: "QualityLayerStyleConfig",
    ) -> None:
        self.style = style

    def find_layer_from_project(self) -> Optional[QgsAnnotationLayer]:
        """
        Find QGIS layer using custom layer id which is automatically
        generated for the layer.
        """
        layers = [
            layer
            for layer in QgsProject.instance().mapLayers().values()
            if isinstance(layer, QgsAnnotationLayer)
            and layer.customProperty(self.LAYER_ID_PROPERTY) == self.LAYER_ID
        ]
        if len(layers) > 1:
            LOGGER.warning(
                f"Found multiple ({len(layers)}) layers with the same identifying "
                f"internal id {self.LAYER_ID}, should have found only one for the "
                "find logic to work on unique instances."
            )
        return layers[0] if len(layers) > 0 else None

    def _create_annotation_layer(self) -> QgsAnnotationLayer:
        layer = QgsAnnotationLayer(
            tr("Quality errors"),
            QgsAnnotationLayer.LayerOptions(QgsProject.instance().transformContext()),
        )

        layer.setCustomProperty(self.LAYER_ID_PROPERTY, self.LAYER_ID)

        LOGGER.debug(f"Created a layer: {layer.name()}")

        return layer

    def get_annotation_layer(self) -> QgsAnnotationLayer:
        """Creates an annotation layer from the layer configuration."""
        layer = self.find_layer_from_project()
        if layer is None:
            layer = self._create_annotation_layer()

        return layer

    def add_or_replace_annotation(
        self,
        quality_error: "QualityError",
        use_highlighted_style: bool,
        id_prefix: str = "",
    ) -> None:
        annotation_layer = self.annotation_layer
        if quality_error.geometry.isNull():
            return

        annotations = self._create_annotations(
            quality_error,
            use_highlighted_style,
        )

        internal_id = f"{id_prefix}{quality_error.unique_identifier}"

        # Update
        if internal_id in self._annotation_ids:
            # Singlepart geometries
            if len(annotations) == 1 and len(self._annotation_ids[internal_id]) == 1:
                annotation_layer.replaceItem(
                    self._annotation_ids[internal_id][0], annotations[0]
                )
            # Multipart geometries
            else:
                for annotation_id in self._annotation_ids[internal_id]:
                    annotation_layer.removeItem(annotation_id)
                new_ids = []
                for annotation in annotations:
                    new_ids.append(annotation_layer.addItem(annotation))
                self._annotation_ids[internal_id] = new_ids
        # New
        else:
            new_ids = []
            for annotation in annotations:
                new_ids.append(annotation_layer.addItem(annotation))
            self._annotation_ids[internal_id] = new_ids

    def remove_annotations(
        self, quality_errors: Iterable["QualityError"], id_prefix: str = ""
    ) -> None:
        annotation_layer = self.annotation_layer

        for quality_error in quality_errors:
            internal_id = f"{id_prefix}{quality_error.unique_identifier}"
            try:
                annotation_ids = self._annotation_ids.pop(internal_id)
                for annotation_id in annotation_ids:
                    annotation_layer.removeItem(annotation_id)
            except KeyError:
                # Consume exception, feature is not found
                pass

    def _create_annotations(
        self,
        quality_error: "QualityError",
        use_highlighted_style: bool,
    ) -> List[
        Union[QgsAnnotationMarkerItem, QgsAnnotationPolygonItem, QgsAnnotationLineItem]
    ]:
        annotations: List[
            Union[
                QgsAnnotationMarkerItem, QgsAnnotationPolygonItem, QgsAnnotationLineItem
            ]
        ] = []
        geometry = quality_error.geometry
        geom_type = geometry.type()

        original_abstract_geometry = geometry.get()
        cloned_original_abstract_geometry = original_abstract_geometry.clone()
        cloned_geom = QgsGeometry(cloned_original_abstract_geometry)

        annotation = None

        symbol = self.style.create_error_symbol(quality_error)

        if geom_type == QgsWkbTypes.PointGeometry:
            points = []
            if geometry.isMultipart() is False:
                point = QgsPoint()
                point.fromWkt(cloned_geom.asWkt())
                points.append(point)
            else:
                for part in cloned_geom.constParts():
                    point = QgsPoint()
                    point.fromWkt(part.asWkt())
                    points.append(point)

            for point in points:
                annotation = QgsAnnotationMarkerItem(point)
                annotation.setSymbol(
                    symbol.to_qgs_symbol(geom_type, use_highlighted_style)
                )
                annotations.append(annotation)

        elif geom_type == QgsWkbTypes.PolygonGeometry:
            polygons = []

            if geometry.isMultipart() is False:
                polygon = QgsPolygon()
                polygon.fromWkt(geometry.asWkt())
                polygons.append(polygon)
            else:
                for part in geometry.constParts():
                    polygon = QgsPolygon()
                    polygon.fromWkt(part.asWkt())
                    polygons.append(polygon)

            for polygon in polygons:
                annotation = QgsAnnotationPolygonItem(polygon)
                annotation.setSymbol(
                    symbol.to_qgs_symbol(geom_type, use_highlighted_style)
                )
                annotations.append(annotation)

        elif geom_type == QgsWkbTypes.LineGeometry:
            lines = []
            if geometry.isMultipart() is False:
                line = QgsLineString()
                line.fromWkt(cloned_geom.asWkt())
                lines.append(line)
            else:
                for part in cloned_geom.constParts():
                    line = QgsLineString()
                    line.fromWkt(part.asWkt())
                    lines.append(line)

            for line in lines:
                annotation = QgsAnnotationLineItem(line)
                annotation.setSymbol(
                    symbol.to_qgs_symbol(geom_type, use_highlighted_style)
                )
                annotations.append(annotation)

        else:
            raise ValueError(f"Unsupported geom type: {geom_type}")

        return annotations



#  Copyright (C) 2022-2023 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import logging
from pathlib import Path
from typing import Optional, cast

import qgis_plugin_tools
from qgis.core import QgsApplication, QgsProject
from qgis.gui import QgisInterface
from qgis.PyQt.QtCore import QCoreApplication, Qt, QTranslator
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction
from qgis.utils import iface as utils_iface
from qgis_plugin_tools.tools.custom_logging import setup_loggers
from qgis_plugin_tools.tools.i18n import setup_translation, tr
from qgis_plugin_tools.tools.resources import resources_path

import quality_result_gui
import quality_result_gui_plugin
from quality_result_gui.quality_error_manager import QualityResultManager
from quality_result_gui_plugin.dev_tools.dev_tools_dialog import DevToolsDialog
from quality_result_gui_plugin.dev_tools.mock_api_client import MockQualityResultClient
from quality_result_gui_plugin.env import IS_DEVELOPMENT_MODE

LOGGER = logging.getLogger(__name__)

iface = cast(QgisInterface, utils_iface)


class QualityResultGuiPlugin:
    def __init__(self) -> None:
        self._teardown_loggers = lambda: None

        self.quality_error_manager: Optional[QualityResultManager] = None

        # Initialize locale
        _, file_path = setup_translation()
        if file_path:
            self.translator = QTranslator()
            self.translator.load(file_path)
            QCoreApplication.installTranslator(self.translator)

        self._menu_name = tr("Quality result GUI")

        self.show_dev_tools_dialog_action: Optional[QAction] = None
        self.dev_tools_dialog: Optional[DevToolsDialog] = None
        self.dev_tool_action: Optional[QAction] = None
        self._test_json_file_path = ""

    def initGui(self) -> None:  # noqa: N802 (qgis naming)
        self._teardown_loggers = setup_loggers(
            quality_result_gui.__name__,
            quality_result_gui_plugin.__name__,
            qgis_plugin_tools.__name__,
            message_log_name=tr("Quality result GUI"),
        )

        # Add action to show dev tools dialog to menu
        self.show_dev_tools_dialog_action = QAction(
            QgsApplication.getThemeIcon("/propertyicons/settings.svg"),
            tr("Development tools"),
            iface.mainWindow(),
        )
        self.show_dev_tools_dialog_action.triggered.connect(
            self._on_open_dev_tools_called
        )

        iface.addPluginToMenu(self._menu_name, self.show_dev_tools_dialog_action)

        # Add shortcut action to open qulity results form example json when in dev mode
        if IS_DEVELOPMENT_MODE and self.dev_tool_action is None:

            self._test_json_file_path = str(
                (
                    Path(__file__).parent
                    / "dev_tools/example_quality_errors/quality_errors.json"
                ).resolve()
            )
            self.dev_tool_action = QAction(
                QIcon(resources_path("icons/quality_result_gui.svg")),
                "Test quality result GUI",
                iface.mainWindow(),
            )
            self.dev_tool_action.triggered.connect(
                self._open_quality_result_gui_in_dev_mode
            )
            iface.addToolBarIcon(self.dev_tool_action)

    def unload(self) -> None:
        iface.removePluginMenu(self._menu_name, self.show_dev_tools_dialog_action)

        self._unload_quality_error_manager_if_exists()

        self._teardown_loggers()
        self._teardown_loggers = lambda: None

        if self.dev_tool_action:
            iface.removeToolBarIcon(self.dev_tool_action)

    def _unload_quality_error_manager_if_exists(self) -> None:
        if self.quality_error_manager:
            iface.removeDockWidget(self.quality_error_manager.dock_widget)

            self.quality_error_manager.unload()
            self.quality_error_manager = None

    def _on_open_dev_tools_called(self) -> None:

        dialog = DevToolsDialog(iface.mainWindow())
        if dialog.exec_():
            self._test_json_file_path = (
                dialog.quality_errors_data_file_widget.filePath()
            )
            self._open_quality_result_gui_in_dev_mode()

    def _open_quality_result_gui_in_dev_mode(self) -> None:
        self._unload_quality_error_manager_if_exists()

        api_client = MockQualityResultClient(Path(self._test_json_file_path))
        QgsProject.instance().setCrs(api_client.get_crs())

        self.quality_error_manager = QualityResultManager(
            api_client, iface.mainWindow()
        )
        self.quality_error_manager.closed.connect(
            self._unload_quality_error_manager_if_exists
        )

        iface.addDockWidget(
            Qt.DockWidgetArea.RightDockWidgetArea,
            self.quality_error_manager.dock_widget,
        )
        self.quality_error_manager.show_dock_widget()



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional

from qgis.core import (
    QgsCoordinateReferenceSystem,
    QgsCoordinateTransform,
    QgsGeometry,
    QgsLayerTree,
    QgsProject,
    QgsRectangle,
    QgsVectorLayer,
)
from qgis.gui import QgisInterface
from qgis.utils import iface as utils_iface

iface: QgisInterface = utils_iface

SUPPORTED_GEOMETRIES = ["MultiPoint", "MultiLineString", "MultiPolygon"]
BOUNDING_BOX_BUFFER_COEFFICIENT = 0.1


def set_visibility_checked(layer: QgsVectorLayer, checked: bool) -> None:
    root: QgsLayerTree = QgsProject.instance().layerTreeRoot()

    tree_node = root.findLayer(layer.id())
    if tree_node is not None:
        tree_node.setItemVisibilityCheckedRecursive(checked)


def zoom_to_geometries_and_flash(
    geometries: List[QgsGeometry],
    crs: QgsCoordinateReferenceSystem,
    preserve_scale: bool = False,
    min_extent_height: Optional[int] = None,
) -> None:
    view_extent = get_extent_from_geometries(geometries)
    if view_extent is None:
        return

    if crs != QgsProject.instance().crs():
        view_extent = transform_bounding_box(
            view_extent, crs, QgsProject.instance().crs()
        )
    # Only move canvas
    if preserve_scale is True:
        iface.mapCanvas().setCenter(view_extent.center())
    # Move canvas and zoom to geometries
    else:
        if min_extent_height is not None and view_extent.height() < min_extent_height:
            view_extent = (
                QgsGeometry.fromPointXY(view_extent.center())
                .buffer(min_extent_height / 2, 1)
                .boundingBox()
            )
        iface.mapCanvas().setExtent(view_extent)
    iface.mapCanvas().flashGeometries(geometries, crs)


def get_extent_from_geometries(geometries: List[QgsGeometry]) -> Optional[QgsRectangle]:
    if len(geometries) == 0:
        return None

    view_extent = geometries[0].buffer(BOUNDING_BOX_BUFFER_COEFFICIENT, 2).boundingBox()
    for geometry in geometries:
        view_extent.combineExtentWith(geometry.boundingBox())

    return view_extent.buffered(view_extent.height() * BOUNDING_BOX_BUFFER_COEFFICIENT)


def transform_bounding_box(
    rectangle: QgsRectangle,
    crs: str,
    target_crs: str,
) -> QgsRectangle:
    """
    Transform bounding box from one crs to other.
    """
    trans = QgsCoordinateTransform(
        QgsCoordinateReferenceSystem(crs),
        QgsCoordinateReferenceSystem(target_crs),
        QgsProject.instance(),
    )
    return trans.transformBoundingBox(rectangle)



#  Copyright (C) 2022 National Land Survey of Finland
#  (https://www.maanmittauslaitos.fi/en).
#
#
#  This file is part of quality-result-gui.
#
#  quality-result-gui is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  quality-result-gui is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with quality-result-gui. If not, see <https://www.gnu.org/licenses/>.

import contextlib
import enum
import logging
from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    NewType,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    overload,
)

from qgis.gui import QgisInterface
from qgis.PyQt.QtCore import (
    QAbstractItemModel,
    QIdentityProxyModel,
    QModelIndex,
    QObject,
    QSortFilterProxyModel,
    Qt,
    QVariant,
    pyqtSignal,
)
from qgis.PyQt.QtGui import QColor, QFont
from qgis.utils import iface as utils_iface
from qgis_plugin_tools.tools.i18n import tr

from quality_result_gui.api.types.quality_error import (
    ERROR_PRIORITY_LABEL,
    ERROR_TYPE_LABEL,
    QualityError,
    QualityErrorPriority,
    QualityErrorsByPriority,
)

if TYPE_CHECKING:
    from qgis.core import QgsRectangle

    from quality_result_gui.quality_errors_filters import AbstractQualityErrorFilter

LOGGER = logging.getLogger(__name__)

iface = cast(QgisInterface, utils_iface)


class ModelColumn(enum.Enum):
    TYPE_OR_ID = 0
    ERROR_DESCRIPTION = 1


COLUMN_HEADERS = {
    ModelColumn.TYPE_OR_ID: tr("Errors"),
    ModelColumn.ERROR_DESCRIPTION: tr("Error description"),
}


def get_error_feature_types(
    errors_by_priority: List[QualityErrorsByPriority],
) -> Set[str]:
    return {
        errors.feature_type
        for errors_by_feature_type in errors_by_priority
        for errors in errors_by_feature_type.errors
    }


def get_error_feature_attributes(
    quality_errors: List[QualityErrorsByPriority],
) -> Set[str]:
    return {
        error.attribute_name
        for errors_by_priority in quality_errors
        for error in errors_by_priority.get_all_errors()
        if error.attribute_name
    }


def _get_quality_errors_indexes(
    model: QAbstractItemModel, index: QModelIndex
) -> Iterator[QModelIndex]:
    """Get quality all error indexes from index."""
    if not index.isValid():
        return

    row_count = model.rowCount(index)
    if row_count == 0:
        data = index.data(Qt.UserRole)
        (item_type, _) = cast(ErrorDataType, data)
        if item_type == QualityErrorTreeItemType.ERROR:
            # Index is for quality error row, which has no children
            yield index
    else:
        for i in range(row_count):
            yield from _get_quality_errors_indexes(model, index.child(i, 0))


def _count_quality_error_rows(model: QAbstractItemModel, index: QModelIndex) -> int:
    if not index.isValid():
        return 0
    num_rows = 0
    row_count = model.rowCount(index)
    if row_count == 0:
        data = index.data(Qt.UserRole)
        (item_type, _) = cast(ErrorDataType, data)
        if item_type == QualityErrorTreeItemType.ERROR:
            # Index is for quality error row
            return 1
    for i in range(row_count):
        child_index = index.child(i, 0)
        num_rows += _count_quality_error_rows(model, child_index)
    return num_rows


def _count_all_rows(model: QAbstractItemModel) -> int:
    num_rows = 0
    for i in range(model.rowCount(QModelIndex())):
        index = model.index(i, 0, QModelIndex())
        num_rows += _count_quality_error_rows(model, index)
    return num_rows


class QualityErrorTreeItemType(enum.Enum):
    HEADER = enum.auto()
    PRIORITY = enum.auto()
    FEATURE_TYPE = enum.auto()
    FEATURE = enum.auto()
    ERROR = enum.auto()


ErrorDataType = NewType(
    "ErrorDataType",
    Tuple[QualityErrorTreeItemType, Any],
)


class QualityErrorTreeItem:
    def __init__(
        self,
        data: List[Any],
        key: str,
        item_type: QualityErrorTreeItemType,
        parent: Optional["QualityErrorTreeItem"] = None,
    ) -> None:
        self.key = key
        self._item_parent = parent
        self._item_data = data
        self._child_items: List["QualityErrorTreeItem"] = []
        self._child_item_map: Dict[str, int] = {}

        self.item_type = item_type

    def append_child_item(self, item: "QualityErrorTreeItem") -> None:
        self._child_item_map[item.key] = len(self._child_items)
        self._child_items.append(item)

    def remove_child_item(self, item: "QualityErrorTreeItem") -> None:
        self._child_items.remove(item)
        self._child_item_map.pop(item.key)

    def get_child_by_key(self, key: str) -> "QualityErrorTreeItem":
        return self._child_items[self._child_item_map[key]]

    def child(self, row: int) -> Optional["QualityErrorTreeItem"]:
        if row >= 0 and row < self.child_count():
            return self._child_items[row]
        return None

    def child_count(self) -> int:
        return len(self._child_items)

    def row(self) -> int:
        if self._item_parent is not None:
            return self._item_parent._child_items.index(self)
        return 0

    def column_count(self) -> int:
        return len(self._item_data)

    def data(
        self, column_index: int, role: Qt.ItemDataRole = Qt.DisplayRole
    ) -> QVariant:
        if not (column_index >= 0 and column_index < len(self._item_data)):
            return QVariant()

        item_data = self._item_data[column_index]
        column_data = QVariant()

        column = ModelColumn(column_index)

        if self.item_type == QualityErrorTreeItemType.HEADER:
            return QVariant()

        if role == Qt.UserRole and column in [
            ModelColumn.TYPE_OR_ID,
            ModelColumn.ERROR_DESCRIPTION,
        ]:
            index = (
                column_index - 1
                if column == ModelColumn.ERROR_DESCRIPTION
                else column_index
            )
            return (self.item_type, self._item_data[index])

        if role == Qt.DisplayRole and column == ModelColumn.TYPE_OR_ID:
            if self.item_type == QualityErrorTreeItemType.PRIORITY:
                column_data = ERROR_PRIORITY_LABEL[QualityErrorPriority(item_data)]

            elif self.item_type == QualityErrorTreeItemType.FEATURE_TYPE:
                column_data = item_data
                # TODO: how to configurate custom data mapping
                # column_data = common.FEATURE_TYPE_NAMES[item_data]

            elif self.item_type == QualityErrorTreeItemType.FEATURE:
                column_data = item_data[1][:8]

            elif self.item_type == QualityErrorTreeItemType.ERROR:
                quality_error = cast(QualityError, item_data)
                column_data = ERROR_TYPE_LABEL[quality_error.error_type]
            return QVariant(column_data)

        if (
            role == Qt.DisplayRole
            and column == ModelColumn.ERROR_DESCRIPTION
            and self.item_type == QualityErrorTreeItemType.ERROR
        ):
            return item_data["error_description"]

        if (
            role == Qt.ToolTipRole
            and column == ModelColumn.ERROR_DESCRIPTION
            and self.item_type == QualityErrorTreeItemType.ERROR
        ):
            extra_info = item_data.get("error_extra_info", None)
            if extra_info is None:
                return item_data["error_description"]
            return f"<p>{item_data['error_description']}</p><p>{extra_info}</p>"

        if (
            role == Qt.CheckStateRole
            and column == ModelColumn.TYPE_OR_ID
            and self.item_type == QualityErrorTreeItemType.ERROR
        ):
            quality_error = cast(QualityError, item_data)
            if quality_error.is_user_processed is True:
                return QVariant(Qt.Checked)
            else:
                return QVariant(Qt.Unchecked)

        return QVariant()

    def parent(self) -> Optional["QualityErrorTreeItem"]:
        return self._item_parent


class QualityErrorsTreeBaseModel(QAbstractItemModel):
    """
    Simple tree model. Adapted to use conflict process directly as the data source.

    https://doc.qt.io/qt-5/qtwidgets-itemviews-simpletreemodel-example.html
    """

    filterable_data_changed = pyqtSignal()
    error_checked = pyqtSignal(str, bool)

    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)

        self._root_item = QualityErrorTreeItem(
            len(COLUMN_HEADERS) * [None],
            "header",
            QualityErrorTreeItemType.HEADER,
        )
        # Show error priority rows always
        for priority in list(QualityErrorPriority):
            priority_item = QualityErrorTreeItem(
                [priority, None],
                str(priority.value),
                QualityErrorTreeItemType.PRIORITY,
                self._root_item,
            )
            self._add_item_to_model(priority_item, self._root_item)

    def index(self, row: int, column: int, parent: QModelIndex) -> QModelIndex:
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parent_item = self._root_item
        else:
            parent_item = parent.internalPointer()

        child = parent_item.child(row)

        if child is not None:
            return self.createIndex(row, column, child)

        return QModelIndex()

    @overload
    def parent(self, child: QModelIndex) -> QModelIndex:
        ...

    @overload
    def parent(self) -> QObject:
        ...

    def parent(
        self, child: Optional[QModelIndex] = None
    ) -> Union[QModelIndex, QObject]:
        if child is None:
            return super().parent()

        if not child.isValid():
            return QModelIndex()

        child_item: QualityErrorTreeItem = child.internalPointer()
        parent_item = child_item.parent()

        if parent_item == self._root_item or parent_item is None:
            return QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def rowCount(self, parent: QModelIndex) -> int:  # noqa: N802 (qt override)
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parent_item = self._root_item
        else:
            parent_item = parent.internalPointer()

        return parent_item.child_count()

    def columnCount(self, parent: QModelIndex) -> int:  # noqa: N802 (qt override)
        if not parent.isValid():
            parent_item = self._root_item
        else:
            parent_item = parent.internalPointer()

        return parent_item.column_count()

    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole
    ) -> QVariant:
        if not index.isValid():
            return QVariant()

        item: QualityErrorTreeItem = index.internalPointer()

        return item.data(index.column(), role)

    def headerData(  # noqa: N802 (qt override)
        self,
        section: int,
        orientation: Qt.Orientation,
        role: Qt.ItemDataRole = Qt.DisplayRole,
    ) -> QVariant:
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            try:
                model_column = ModelColumn(section)
                if model_column == ModelColumn.TYPE_OR_ID:
                    total_count = _count_all_rows(self)
                    return QVariant(
                        f"{COLUMN_HEADERS.get(model_column, QVariant())}"
                        f" ({total_count}/{total_count})"
                    )
                return COLUMN_HEADERS.get(model_column, QVariant())
            except ValueError:
                return QVariant()
        return QVariant()

    def setData(  # noqa: N802 (qt override)
        self, index: QModelIndex, value: Any, role: Qt.ItemDataRole = Qt.EditRole
    ) -> bool:
        if not index.isValid() or role == Qt.EditRole:
            return False

        column = ModelColumn(index.column())
        item: QualityErrorTreeItem = index.internalPointer()

        if (
            column == ModelColumn.TYPE_OR_ID
            and item.item_type == QualityErrorTreeItemType.ERROR
        ):
            quality_error = cast(QualityError, item._item_data[0])

            if role == Qt.CheckStateRole:
                checked = value == Qt.Checked
                quality_error.is_user_processed = checked

                self.error_checked.emit(quality_error.unique_identifier, checked)
                self.dataChanged.emit(index, index)

                parent_index = index
                while parent_index.isValid():
                    parent_index = parent_index.parent()
                    self.dataChanged.emit(parent_index, parent_index)
                return True

        return False

    def flags(  # (qt override)
        self,
        index: QModelIndex,
    ) -> Qt.ItemFlags:
        if not index.isValid():
            return Qt.NoItemFlags

        column = ModelColumn(index.column())
        item: QualityErrorTreeItem = index.internalPointer()

        if (
            column == ModelColumn.TYPE_OR_ID
            and item.item_type == QualityErrorTreeItemType.ERROR
        ):
            return super().flags(index) | Qt.ItemIsUserCheckable

        return super().flags(index)

    def refresh_model(self, quality_errors: List[QualityErrorsByPriority]) -> None:

        updated_quality_error_ids = {
            error.unique_identifier
            for errors_by_priority in quality_errors
            for error in errors_by_priority.get_all_errors()
        }

        current_quality_error_ids = set()
        for i in range(self.rowCount(QModelIndex())):
            for index in _get_quality_errors_indexes(
                self, self.index(i, 0, QModelIndex())
            ):
                _, item_data = index.data(Qt.UserRole)
                current_quality_error_ids.add(
                    cast(QualityError, item_data).unique_identifier
                )

        deleted_error_ids = current_quality_error_ids - updated_quality_error_ids
        new_error_ids = updated_quality_error_ids - current_quality_error_ids

        # Nothing changed
        if not deleted_error_ids and not new_error_ids:
            return

        errors_to_be_added = (
            error
            for errors_by_priority in quality_errors
            for error in errors_by_priority.get_all_errors()
            if error.unique_identifier in new_error_ids
        )

        errors_to_be_deleted: List[Tuple[QualityErrorTreeItem, QModelIndex]] = []

        for i in range(self.rowCount(QModelIndex())):
            for index in _get_quality_errors_indexes(
                self, self.index(i, 0, QModelIndex())
            ):
                item: QualityErrorTreeItem = index.internalPointer()
                if item.key in deleted_error_ids:
                    errors_to_be_deleted.append((item, index))

        self._update_model_data(errors_to_be_added, errors_to_be_deleted)

        self.filterable_data_changed.emit()

    def _update_model_data(
        self,
        errors_to_be_added: Iterable[QualityError],
        errors_to_be_deleted: List[Tuple[QualityErrorTreeItem, QModelIndex]],
    ) -> None:
        """
        Updates model data based on new and deleted quality errors.

        Deletion of quality errors must be done in reversed order for model
        indices to stay valid during the update process. New items are added
        after deletion. Empty parents are left to the model as filter model
        will leave them out eventually.
        """
        # Remove quality error items that are no longer found from errors
        for item, item_index in reversed(errors_to_be_deleted):
            self._remove_item_from_model(item, item_index)

        # Add new quality error items and parent items for them if needed
        for quality_error in errors_to_be_added:
            priority_item = self._root_item.get_child_by_key(
                str(quality_error.priority.value)
            )

            try:
                feature_type_item = priority_item.get_child_by_key(
                    quality_error.feature_type
                )
            except KeyError:
                feature_type_item = QualityErrorTreeItem(
                    [quality_error.feature_type, None],
                    quality_error.feature_type,
                    QualityErrorTreeItemType.FEATURE_TYPE,
                    priority_item,
                )
                self._add_item_to_model(
                    feature_type_item,
                    priority_item,
                )

            try:
                feature_item = feature_type_item.get_child_by_key(
                    quality_error.feature_id
                )
            except KeyError:
                feature_item = QualityErrorTreeItem(
                    [
                        (
                            quality_error.feature_type,
                            quality_error.feature_id,
                        ),
                        None,
                    ],
                    quality_error.feature_id,
                    QualityErrorTreeItemType.FEATURE,
                    feature_type_item,
                )
                self._add_item_to_model(
                    feature_item,
                    feature_type_item,
                )

            quality_error_item = QualityErrorTreeItem(
                [
                    quality_error,
                    {
                        "error_description": quality_error["error_description"],
                        "error_extra_info": quality_error["error_extra_info"],
                    },
                ],
                quality_error.unique_identifier,
                QualityErrorTreeItemType.ERROR,
                feature_item,
            )

            self._add_item_to_model(
                quality_error_item,
                feature_item,
            )

    def _get_index_for_item(self, item: QualityErrorTreeItem) -> QModelIndex:

        if item.item_type == QualityErrorTreeItemType.HEADER:
            return QModelIndex()
        else:
            item_parent = item.parent()
            if item_parent is None:
                raise ValueError
            return self.index(item.row(), 0, self._get_index_for_item(item_parent))

    def _remove_item_from_model(
        self,
        item: QualityErrorTreeItem,
        item_index: QModelIndex,
    ) -> None:
        item_parent = item.parent()
        if item_parent is None:
            return

        self.beginRemoveRows(item_index.parent(), item.row(), item.row())
        item_parent.remove_child_item(item)
        self.endRemoveRows()

    def _add_item_to_model(
        self, item: QualityErrorTreeItem, item_parent: QualityErrorTreeItem
    ) -> None:
        parent_index = self._get_index_for_item(item_parent)
        self.beginInsertRows(
            parent_index,
            item_parent.child_count(),
            item_parent.child_count(),
        )
        item_parent.append_child_item(item)
        self.endInsertRows()


class StyleProxyModel(QIdentityProxyModel):
    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)

    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole
    ) -> QVariant:
        source_index = self.mapToSource(index)
        data = self.sourceModel().data(source_index, Qt.UserRole)

        if not QVariant(data).isValid():
            return self.sourceModel().data(source_index, role)

        (item_type, item_data) = cast(ErrorDataType, data)
        column = ModelColumn(index.column())
        if (
            role == Qt.FontRole
            and column == ModelColumn.TYPE_OR_ID
            and item_type == QualityErrorTreeItemType.PRIORITY
        ):
            font = QFont()
            font.setBold(True)
            font.setPointSize(10)
            return QVariant(font)

        if (
            role == Qt.FontRole
            and column == ModelColumn.TYPE_OR_ID
            and item_type == QualityErrorTreeItemType.FEATURE_TYPE
        ):
            font = QFont()
            font.setBold(True)
            return QVariant(font)

        if (
            role == Qt.ForegroundRole
            and column == ModelColumn.TYPE_OR_ID
            and item_type == QualityErrorTreeItemType.ERROR
            and cast(QualityError, item_data).is_user_processed is True
        ):
            return QVariant(QColor(Qt.lightGray))

        if (
            role == Qt.ForegroundRole
            and column == ModelColumn.TYPE_OR_ID
            and item_type == QualityErrorTreeItemType.FEATURE
        ):
            color = QColor()
            color.setRgb(75, 75, 75)
            return QVariant(color)

        return self.sourceModel().data(source_index, role)


class AbstractFilterProxyModel(QSortFilterProxyModel):
    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)
        self.setFilterRole(Qt.UserRole)

    def filterAcceptsRow(  # noqa: N802 (qt override)
        self, source_row: int, source_parent: QModelIndex
    ) -> bool:
        source_index = self.sourceModel().index(source_row, 0, source_parent)

        if not source_index.isValid():
            return True

        data = self.sourceModel().data(source_index, self.filterRole())
        if not QVariant(data).isValid():
            # Always accept anything that did not return valid data
            return True

        (item_type, item_value) = cast(ErrorDataType, data)

        row_accepted = self.accept_row(item_type, item_value)
        if not row_accepted:
            return False

        if item_type == QualityErrorTreeItemType.ERROR:
            return True
        else:
            return self._is_any_children_visible(source_index)

    @abstractmethod
    def accept_row(
        self, tree_item_type: QualityErrorTreeItemType, tree_item_value: Any
    ) -> bool:
        raise NotImplementedError()

    def _is_any_children_visible(self, source_index: QModelIndex) -> bool:
        child_count = self.sourceModel().rowCount(source_index)
        return any(
            self.filterAcceptsRow(child_row_num, source_index)
            for child_row_num in range(child_count)
        )


class FilterProxyModel(AbstractFilterProxyModel):
    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)

        self._filters: List["AbstractQualityErrorFilter"] = []

    def add_filter(self, filter: "AbstractQualityErrorFilter") -> None:
        filter.filters_changed.connect(self.invalidateFilter)
        self._filters.append(filter)

        self.invalidateFilter()

    def accept_row(self, item_type: QualityErrorTreeItemType, item_value: Any) -> bool:
        # TODO: Check only the changed filters.
        # Now this checks all the filters when one changes
        return all(
            quality_filter.accept_row(item_type, item_value)
            for quality_filter in self._filters
        )

    def headerData(  # noqa: N802 (qt override)
        self,
        section: int,
        orientation: Qt.Orientation,
        role: Qt.ItemDataRole = Qt.DisplayRole,
    ) -> QVariant:
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            try:
                model_column = ModelColumn(section)
                if model_column == ModelColumn.TYPE_OR_ID:
                    total_count = (
                        self.sourceModel()
                        .headerData(section, orientation, role)
                        .value()
                        .split("/")[1]
                    )
                    total_count = "".join([n for n in total_count if n.isdigit()])

                    filtered_count = _count_all_rows(self)
                    return QVariant(
                        f"{COLUMN_HEADERS.get(model_column, QVariant())}"
                        f" ({filtered_count}/{total_count})"
                    )
                return COLUMN_HEADERS.get(model_column, QVariant())
            except ValueError:
                return QVariant()
        return QVariant()


class FilterByExtentProxyModel(AbstractFilterProxyModel):
    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)

        self._extent: Union["QgsRectangle", None] = None

    def set_extent(self, extent: Union["QgsRectangle", None]) -> None:
        self._extent = extent
        self.invalidateFilter()

    def _on_map_extent_changed(self) -> None:
        self.set_extent(iface.mapCanvas().extent())

    def set_enabled(self, enabled: bool) -> None:
        if enabled:
            iface.mapCanvas().extentsChanged.connect(self._on_map_extent_changed)
            self.set_extent(iface.mapCanvas().extent())
        else:
            with contextlib.suppress(TypeError):  # Ignore case when not connected
                iface.mapCanvas().extentsChanged.disconnect(self._on_map_extent_changed)

            self.set_extent(None)

    def accept_row(
        self, tree_item_type: QualityErrorTreeItemType, tree_item_value: Any
    ) -> bool:
        if not self._extent:
            return True
        if tree_item_type != tree_item_type.ERROR:
            return True

        quality_error = cast(QualityError, tree_item_value)

        return quality_error.geometry.intersects(self._extent)


class FilterByShowUserProcessedProxyModel(AbstractFilterProxyModel):
    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)

        self._show_processed_errors: bool = True

    def set_show_processed_errors(self, show_processed_errors: bool) -> None:
        self._show_processed_errors = show_processed_errors
        self.invalidateFilter()

    def accept_row(
        self, tree_item_type: QualityErrorTreeItemType, tree_item_value: Any
    ) -> bool:
        return not (
            self._show_processed_errors is False
            and tree_item_type == QualityErrorTreeItemType.ERROR
            and cast(QualityError, tree_item_value).is_user_processed is True
        )
